import rclpy  # ROS client library
from rclpy.node import Node
from rclpy.qos import qos_profile_sensor_data

from sensor_msgs.msg import LaserScan
from geometry_msgs.msg import Twist

import time

from datetime import datetime
from datetime import timedelta


class Tb3(Node):
    def __init__(self):
        super().__init__('tb3')

        self.cmd_vel_pub = self.create_publisher(
                Twist,      # message type
                'cmd_vel',  # topic name
                1)          # history depth

        self.scan_sub = self.create_subscription(
                LaserScan,
                'scan',
                self.scan_callback,  # function to run upon message arrival
                qos_profile_sensor_data)  # allows packet loss

        self.ang_vel_percent = 0
        self.lin_vel_percent = 0
        self.safe_distance = 0.4
        self.turn_allowed = False
        self.stop = False
        self.timestamp = 0
    
    def vel(self, lin_vel_percent, ang_vel_percent=0):
        """ publishes linear and angular velocities in percent
        """
        # for TB3 Waffle
        MAX_LIN_VEL = 0.26  # m/s
        MAX_ANG_VEL = 1.82  # rad/s

        cmd_vel_msg = Twist()
        cmd_vel_msg.linear.x = MAX_LIN_VEL * lin_vel_percent / 100
        cmd_vel_msg.angular.z = MAX_ANG_VEL * ang_vel_percent / 100

        self.cmd_vel_pub.publish(cmd_vel_msg)
        self.ang_vel_percent = ang_vel_percent
        self.lin_vel_percent = lin_vel_percent

    def drive(self, msg):
        if not self.stop:
            if not self.turn_allowed:
                if msg.ranges[0] > self.safe_distance or msg.ranges[0] == 0.0:
                    print(msg.intensities[0])
                    self.vel(50, 0)
                else:
                    self.turn_allowed = True
                    self.timestamp = time.time()
        
            if (time.time() - self.timestamp) <= 1 and self.turn_allowed:
                self.vel(0, 45)
            if (time.time() - self.timestamp) > 2 and self.turn_allowed:
                self.vel(0, 0)
                self.turn_allowed = False

    def scan_callback(self, msg):
        """ is run whenever a LaserScan msg is received
        """

        self.drive(msg)
        # self.detect_left_right(msg)

        print('Distances:')
        print('⬆️ :', msg.ranges[0], msg.intensities[0])
        print('⬇️ :', msg.ranges[180], msg.intensities[180])
        print('⬅️ :', msg.ranges[90], msg.intensities[90])
        print('➡️ :', msg.ranges[-90], msg.intensities[-90])


def main(args=None):
    rclpy.init(args=args)

    tb3 = Tb3()
    print('waiting for messages...')

    try:
        rclpy.spin(tb3)  # Execute tb3 node
        # Blocks until the executor (spin) cannot work
    except KeyboardInterrupt:
        pass

    tb3.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()