import rclpy  # ROS client library
from rclpy.node import Node
from rclpy.qos import qos_profile_sensor_data

from sensor_msgs.msg import LaserScan
from geometry_msgs.msg import Twist

import time

from datetime import datetime
from datetime import timedelta


class Tb3(Node):
    def __init__(self):
        super().__init__('tb3')

        self.cmd_vel_pub = self.create_publisher(
                Twist,      # message type
                'cmd_vel',  # topic name
                1)          # history depth

        self.scan_sub = self.create_subscription(
                LaserScan,
                'scan',
                self.scan_callback,  # function to run upon message arrival
                qos_profile_sensor_data)  # allows packet loss

        self.ang_vel_percent = 0
        self.lin_vel_percent = 0
        self.safe_distance = 0.4
        self.turn_allowed = False
        self.stop = False
        self.timestamp = 0
    
    def vel(self, lin_vel_percent, ang_vel_percent=0):
        """ publishes linear and angular velocities in percent
        """
        # for TB3 Waffle
        MAX_LIN_VEL = 0.26  # m/s
        MAX_ANG_VEL = 1.82  # rad/s

        cmd_vel_msg = Twist()
        cmd_vel_msg.linear.x = MAX_LIN_VEL * lin_vel_percent / 100
        cmd_vel_msg.angular.z = MAX_ANG_VEL * ang_vel_percent / 100

        self.cmd_vel_pub.publish(cmd_vel_msg)
        self.ang_vel_percent = ang_vel_percent
        self.lin_vel_percent = lin_vel_percent

    def drive(self, msg):
        front = msg.ranges[0]
        left = msg.ranges[90]
        right = msg.ranges[-90]
        rear = msg.ranges[180]
        front_area = msg.ranges[-40:40]
        left_white_distance = msg.ranges[40]
        right_white_distance = msg.ranges[-40]

        if not self.stop:
            if not self.turn_allowed:
                if (front > self.safe_distance or front == 0.0) \
                and (left_white_distance > self.safe_distance or left_white_distance == 0.0) \
                and (right_white_distance > self.safe_distance or right_white_distance == 0.0):
                    self.vel(80, 0)
                else:
                    # white
                    if msg.intensities[0] == 2.0:
                        self.vel(0, 0)
                        self.stop = False
                    else:
                        self.turn_allowed = True
                        self.timestamp = time.time()

            if self.turn_allowed:
                if left < right:
                    self.vel(0, -80)
                    if self.check_to_stop_turn(msg):
                        self.vel(0, 0)
                        self.turn_allowed = False
                else:
                    self.vel(0, 80)
                    if self.check_to_stop_turn(msg):
                        self.vel(0, 0)
                        self.turn_allowed = False
                
    def check_to_stop_turn(self, msg):
        if msg.ranges[0] > 1.2 and msg.ranges[90] > 0.2 and msg.ranges[-90] > 0.2 \
        and msg.ranges[40] > self.safe_distance and msg.ranges[-40] > self.safe_distance:
            return True
        else:
            return False

    def scan_callback(self, msg):
        """ is run whenever a LaserScan msg is received
        """

        self.drive(msg)
        # self.detect_left_right(msg)

        print('Distances:')
        print('⬆️ :', msg.ranges[0], msg.intensities[0])
        print('⬇️ :', msg.ranges[180], msg.intensities[180])
        print('⬅️ :', msg.ranges[90], msg.intensities[90])
        print('➡️ :', msg.ranges[-90], msg.intensities[-90])


def main(args=None):
    rclpy.init(args=args)

    tb3 = Tb3()
    print('waiting for messages...')

    try:
        rclpy.spin(tb3)  # Execute tb3 node
        # Blocks until the executor (spin) cannot work
    except KeyboardInterrupt:
        pass

    tb3.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()